% Fit impedance spectroscopy data using a solar cell model containing 
% PN junction, LH junction, series resistance, and inductor

%% load impedance data and set initial values
clc
clear all

%Load data and set initial series resistance
impedancedata = xlsread('500mV.xlsx'); %Load impedance spectroscopy data
Rser_LOANA = 8.82; %initial series resistance (mOhm) from LOANA IV measurement 
%LOANA Rs values are: Al-BSF-1 8.82; Al-BSF-2 4.50; Al-BSF-3 6.75; PERC 4.33; TOPCon-1 5.71; TOPCon-2 5.16; IBC 5.57; SHJ 3.84       
L_init_set = 100 ; %Inductance (nH)
C_init_set = 10;   %Capacitance (uF) 

frequency_full = impedancedata(:,1); % in Hz
omega_full = 2*pi*frequency_full;
Mag_full = impedancedata(:,2); 
Angle_full = impedancedata(:,3); 
real_full = Mag_full.*cosd(Angle_full);
complex_full = Mag_full.*sind(Angle_full);

%% global fitting for initial values
% impedance model contains a PN junction, series resistance, and inductor

% set initial value, lower bound, and upper bound for all the fitting parameters 

% PN junction resistance
RD_init = Mag_full(1);
RD_low = 0; %Ohm, lower bound
RD_high = Inf; %Ohm, higher bound

% Series resistance
Rser = Rser_LOANA; %Rser_LOANA; %mOhm, fill in value from LOANA
Rser_low = 0.5*Rser_LOANA; %mOhm, lower bound
Rser_high = 2*Rser_LOANA; %mOhm, upper bound

% PN junction capacitance 
C_init = C_init_set; % in uF 
C_low = 0; % in uF, lower bound
C_high = Inf; % in uF, upper bound

%inductance 
L_init = L_init_set; %L_init_set ; %nH
L_low = 0 ; % in nH, lower bound
L_high = Inf; %in nH, upper bound

omegafit = [omega_full omega_full];
impedancefit = [real_full complex_full];

fun1 = @(fit,omegafit) fit(1) + fit(2)./(1+(fit(2)^2)*(omegafit.^2).*(fit(3).^2));
fun2 = @(fit,omegafit) omegafit.*fit(4)-((fit(2)^2)*omegafit.*fit(3))./(1+(fit(2)^2)*(omegafit.^2).*(fit(3).^2));

fun = @(fit,omegafit) [fun1(fit,omegafit(:,1)),  fun2(fit,omegafit(:,2))];

fit0 = [Rser/1000 RD_init C_init*10^-6 L_init*10^-9];
lb = [Rser_low/1000 RD_low C_low*10^-6 L_low*10^-9];
ub = [Rser_high/1000 RD_high C_high*10^-6 L_high*10^-9];

fit = lsqcurvefit(fun,fit0,omegafit,impedancefit,lb,ub);
Rs_global = fit(1); RD_global = fit(2); C_global = fit(3); L_global = fit(4);

%% Fitting with final model including LH junction
% impedance model contains a PN junction, LH junction, series resistance, and inductor

%R_LH upper and lower bounds
R_LH_init = 0.02*RD_global;    % in mOhm, initial  
R_LH_low = 0;            % in mOhm, lower bound
R_LH_high = Inf;            % in mOhm, upper bound

%C_LH upper and lower bounds
C_LH_init = 0.2*C_global;        % in F, initial  
C_LH_low = 0;         % in uF, lower bound
C_LH_high = Inf; % in uF, upper bound

omegafit = [omega_full omega_full];
impedancefit = [real_full complex_full];

fun1 = @(fit,omegafit) fit(1)+ (fit(2)./(1+(fit(2)^2)*(omegafit.^2).*(fit(3).^2))) + (fit(4)./(1+(fit(4)^2)*(omegafit.^2).*(fit(5).^2)));
fun2 = @(fit,omegafit) omegafit.*fit(6)-((fit(2)^2)*omegafit.*fit(3))./(1+(fit(2)^2)*(omegafit.^2).*(fit(3).^2))-((fit(4)^2)*omegafit.*fit(5))./(1+(fit(4)^2)*(omegafit.^2).*(fit(5).^2));

fun = @(fit,omegafit) [fun1(fit,omegafit(:,1)),  fun2(fit,omegafit(:,2))];

fit0 = [Rser/1000 RD_global C_global R_LH_init C_LH_init L_init*10^-9];
lb = [Rser_low/1000 RD_low C_low*10^-6 R_LH_low*10^-3 C_LH_low*10^-6 L_low*10^-9];
ub = [Rser_high/1000 RD_high C_high*10^-6 R_LH_high*10^-3 C_LH_high L_high*10^-9];

fit = lsqcurvefit(fun,fit0,omegafit,impedancefit,lb,ub);
Rs_fit = fit(1); Rj_fit = fit(2); Cj_fit = fit(3); R_LH_fit = fit(4); C_LH_fit = fit(5); L_fit = fit(6); 

disp(['CNLS best-fit values are Cj = ',num2str(Cj_fit*10^6),' uF, ', ...
    'Rj = ',num2str(Rj_fit),' Ohm, ',...
    'C_LH = ',num2str(C_LH_fit*10^6),' uF, ', ...
    'R_LH = ',num2str(R_LH_fit),' Ohm, ',...
    'Rs = ',num2str(Rs_fit*10^3),' mOhm',...
    'L = ',num2str(L_fit*10^9),' nH ']);

%% Fitting quality assessment

Real_fit_full = fit(1)+ (fit(2)./(1+(fit(2)^2)*(omega_full.^2).*(fit(3).^2))) + (fit(4)./(1+(fit(4)^2)*(omega_full.^2).*(fit(5).^2)));
Compl_fit_full = omega_full.*fit(6)-((fit(2)^2)*omega_full.*fit(3))./(1+(fit(2)^2)*(omega_full.^2).*(fit(3).^2))-((fit(4)^2)*omega_full.*fit(5))./(1+(fit(4)^2)*(omega_full.^2).*(fit(5).^2)); 
Mag_fit_full = sqrt(Real_fit_full.^2+Compl_fit_full.^2);
Angle_fit_full = atand(Compl_fit_full./Real_fit_full);

%calculate NRMSE magnitude
MSE_mag = goodnessOfFit(Mag_full,Mag_fit_full,'MSE');
RMSE_mag = sqrt(MSE_mag);
NRMSE_mag = abs(RMSE_mag/(max(Mag_full)-min(Mag_full)));

%calculate RMSE phase
MSE_phase = goodnessOfFit(Angle_full,Angle_fit_full,'MSE');
RMSE_phase = sqrt(MSE_phase);
NRMSE_phase = abs(RMSE_phase/(max(Angle_full)-min(Angle_full)));

%calculate NRMSE resistance
MSE_real = goodnessOfFit(real_full,Real_fit_full,'MSE');
RMSE_real = sqrt(MSE_real);
NRMSE_real = abs(RMSE_real/(max(real_full)-min(real_full)));

%calculate NRMSE reactance
MSE_complex = goodnessOfFit(complex_full,Compl_fit_full,'MSE');
RMSE_complex = sqrt(MSE_complex);
NRMSE_complex = abs(RMSE_complex/(max(complex_full)-min(complex_full)));

disp(['NRMSE M = ',num2str(NRMSE_mag),' , ', ...
    'NRMSE ph = ',num2str(RMSE_phase),' , ',...
    'NRMSE R = ',num2str(NRMSE_real),' , ', ...
    'NRMSE X = ',num2str(NRMSE_complex)]);

%% plot raw data and fitted data combined polar

%fitting data
frequencyplot = linspace(min(frequency_full),max(frequency_full),100000);
omegaplot = linspace(min(omega_full),max(omega_full),100000);
Realfit = fit(1)+ (fit(2)./(1+(fit(2)^2)*(omegaplot.^2).*(fit(3).^2))) + (fit(4)./(1+(fit(4)^2)*(omegaplot.^2).*(fit(5).^2)));
Complfit = omegaplot.*fit(6)-((fit(2)^2)*omegaplot.*fit(3))./(1+(fit(2)^2)*(omegaplot.^2).*(fit(3).^2))-((fit(4)^2)*omegaplot.*fit(5))./(1+(fit(4)^2)*(omegaplot.^2).*(fit(5).^2)); 
Magfit = sqrt(Realfit.^2+Complfit.^2);
Anglefit = atand(Complfit./Realfit);

close all

figure
plot(real_full,-complex_full,'o','MarkerSize',9,'Color','#7570b3')
hold on
plot(Realfit,-Complfit,'--','LineWidth',1, 'Color' , '#d95f02')
grid on
xlabel('R (\Omega)','FontSize',14) 
ylabel('-X(\Omega)','FontSize',14)
legend('Data','Fit','Location','southeast','FontSize',10)


