% Fit impedance spectroscopy data using a solar cell model containing 
% only PN junction 

%% load impedance data and set initial values
clc
clear all

%Load data and set initial series resistance
impedancedata = xlsread('000mV.xlsx'); %Load impedance spectroscopy data
C_init_set = 10;   %Capacitance (uF) 

frequency_full = impedancedata(:,1); % in Hz
omega_full = 2*pi*frequency_full;
Mag_full = impedancedata(:,2); 
Angle_full = impedancedata(:,3); 
real_full = Mag_full.*cosd(Angle_full);
complex_full = Mag_full.*sind(Angle_full);

%% global fitting for initial values
% impedance model contains a PN junction, series resistance, and inductor

% set initial value, lower bound, and upper bound for all the fitting parameters 

% PN junction resistance
RD_init = Mag_full(1); 
RD_low = 0; %Ohm, lower bound
RD_high = Inf; %Ohm, higher bound

% PN junction capacitance 
C_init = C_init_set; % in uF 
C_low = 0; % in uF, lower bound
C_high = Inf; % in uF, upper bound

omegafit = [omega_full omega_full];
impedancefit = [real_full complex_full];

fun1 = @(fit,omegafit) fit(1)./(1+(fit(1)^2)*(omegafit.^2).*(fit(2).^2));
fun2 = @(fit,omegafit) -((fit(1)^2)*omegafit.*fit(2))./(1+(fit(1)^2)*(omegafit.^2).*(fit(2).^2));

fun = @(fit,omegafit) [fun1(fit,omegafit(:,1)),  fun2(fit,omegafit(:,2))];

fit0 = [RD_init C_init*10^-6];
lb = [RD_low C_low];
ub = [RD_high C_high];

fit = lsqcurvefit(fun,fit0,omegafit,impedancefit,lb,ub);
RD_global = fit(1); C_global = fit(2); 

%% PN junction fitting

omegafit = [omega_full omega_full];
impedancefit = [real_full complex_full];

fun1 = @(fit,omegafit) fit(1)./(1+(fit(1)^2)*(omegafit.^2).*(fit(2).^2));
fun2 = @(fit,omegafit) -((fit(1)^2)*omegafit.*fit(2))./(1+(fit(1)^2)*(omegafit.^2).*(fit(2).^2));

fun = @(fit,omegafit) [fun1(fit,omegafit(:,1)),  fun2(fit,omegafit(:,2))];

fit0 = [RD_global C_global];
lb = [RD_low C_low];
ub = [RD_high C_high];

fit = lsqcurvefit(fun,fit0,omegafit,impedancefit,lb,ub);
Rj_fit = fit(1); Cj_fit = fit(2);

disp(['CNLS best-fit values are Cj = ',num2str(Cj_fit*10^6),' uF, ', ...
     'Rj = ',num2str(Rj_fit),' Ohm ']);


%% Determine global goodness of fit

Real_fit_full = fit(1)./(1+(fit(1)^2)*(omega_full.^2).*(fit(2).^2));
Compl_fit_full = -((fit(1)^2)*omega_full.*fit(2))./(1+(fit(1)^2)*(omega_full.^2).*(fit(2).^2));
Mag_fit_full = sqrt(Real_fit_full.^2+Compl_fit_full.^2);
Angle_fit_full = atand(Compl_fit_full./Real_fit_full);

%calculate NRMSE magnitude
MSE_mag = goodnessOfFit(Mag_full,Mag_fit_full,'MSE');
RMSE_mag = sqrt(MSE_mag);
NRMSE_mag = abs(RMSE_mag/(max(Mag_full)-min(Mag_full)));

%calculate RMSE phase
MSE_phase = goodnessOfFit(Angle_full,Angle_fit_full,'MSE');
RMSE_phase = sqrt(MSE_phase);
NRMSE_phase = abs(RMSE_phase/(max(Angle_full)-min(Angle_full)));

%calculate NRMSE resistance
MSE_real = goodnessOfFit(real_full,Real_fit_full,'MSE');
RMSE_real = sqrt(MSE_real);
NRMSE_real = abs(RMSE_real/(max(real_full)-min(real_full)));

%calculate NRMSE reactance
MSE_complex = goodnessOfFit(complex_full,Compl_fit_full,'MSE');
RMSE_complex = sqrt(MSE_complex);
NRMSE_complex = abs(RMSE_complex/(max(complex_full)-min(complex_full)));

disp(['NRMSE M = ',num2str(NRMSE_mag),' , ', ...
    'NRMSE ph = ',num2str(RMSE_phase),' , ',...
    'NRMSE R = ',num2str(NRMSE_real),' , ', ...
    'NRMSE X = ',num2str(NRMSE_complex)]);

%% plot raw data and fitted data combined polar

%fitting data
frequencyplot = linspace(min(frequency_full),max(frequency_full),100000);
omegaplot = linspace(min(omega_full),max(omega_full),100000);
Realfit = fit(1)./(1+(fit(1)^2)*(omegaplot.^2).*(fit(2).^2));
Complfit = -((fit(1)^2)*omegaplot.*fit(2))./(1+(fit(1)^2)*(omegaplot.^2).*(fit(2).^2));
Magfit = sqrt(Realfit.^2+Complfit.^2);
Anglefit = atand(Complfit./Realfit);

close all

figure
plot(real_full,-complex_full,'o','MarkerSize',9,'Color','#7570b3')
hold on
plot(Realfit,-Complfit,'--','LineWidth',1, 'Color' , '#d95f02')
grid on
xlabel('R (\Omega)','FontSize',14) 
ylabel('-X(\Omega)','FontSize',14)
legend('Data','Fit','Location','southeast','FontSize',10)
